/*
Copyright (c) 2005 by Juliusz Chroboczek

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "convert.h"
#include "pluto.h"
#include "sys.h"

int localeEncoding = ISO8859_1;

static char scratch[BUFFER_SIZE + 6];

static int utf8Bits(char c);
static int write_stubbornly(int fd, char *buf, int count);
static int l2u(char*, char*);
static int u2u(char*, char*);

int
copyOut(int fd, char *buf, int count, int force)
{
    int i = 0, j = 0, k, n;

    while(i < count) {
        if(BUFFER_SIZE - j < 0) {
            write_stubbornly(fd, scratch, j);
            j = 0;
        }
        n = guess ? utf8Bits(buf[i]) : 1;

        if(n <= 1) {
            k = l2u(scratch + j, buf + i);
            if(k <= 0)
                abort();
            i++;
            j+= k;
        }
        else if(i + n > count) {
            if(force) {
                k = l2u(scratch + j, buf + i);
                if(k <= 0)
                    abort();
                i++;
                j += k;
            } else {
                if(j > 0)
                    write_stubbornly(fd, scratch, j);
                memmove(buf, buf + i, count - i);
                return count - i;
            }
        } else {
            k = u2u(scratch + j, buf + i);
            if(k > 0) {
                i += k;
                j += k;
            } else {
                j += l2u(scratch + j, buf + i);
                i++;
            }
        }
    }
    if(j > 0)
        write_stubbornly(fd, scratch, j);
    return 0;
}

int
copyIn(int fd, char *buf, int count)
{
    write_stubbornly(fd, buf, count);
    return 0;
}

static int
write_stubbornly(int fd, char *buf, int n)
{
    int i = 0;
    int rc;
    while(i < n) {
        rc = write(fd, buf + i, n - i);
        if(rc < 0 && (errno == EINTR || errno == EAGAIN))
            rc = 0;
        if(rc < 0)
            return (i > 0 ? i : rc);
        i += rc;
        if(i < n)
            waitForOutput(fd);
    }
    return n;
}

static int
utf8Bits(char c)
{
    if((c & 0x80) == 0)
        return 0;
    else if((c & 0xC0) == 0x80)
        return 1;
    else if((c & 0xE0) == 0xC0)
        return 2;
    else if((c & 0xF0) == 0xE0)
        return 3;
    else if((c & 0xF8) == 0xF0)
        return 4;
    else
        return 0;
}


static int 
u2u(char *d, char *s)
{
    int i;
    int n = utf8Bits(*s);
    if(n == 0) {
        *d++ = *s;
        return 1;
    } else if(n == 1) {
        return 0;
    } else {
        for(i = 1; i < n; i++)
            if((s[i] & 0xC0) != 0x80)
                return 0;
        for(i = 0; i < n; i++)
            *d++ = *s++;
        return n;
    }
}

static unsigned short iso8859_2_tophalf[]=
{ 0x00A0, 0x0104, 0x02D8, 0x0141, 0x00A4, 0x013D, 0x015A, 0x00A7,
  0x00A8, 0x0160, 0x015E, 0x0164, 0x0179, 0x00AD, 0x017D, 0x017B,
  0x00B0, 0x0105, 0x02DB, 0x0142, 0x00B4, 0x013E, 0x015B, 0x02C7,
  0x00B8, 0x0161, 0x015F, 0x0165, 0x017A, 0x02DD, 0x017E, 0x017C,
  0x0154, 0x00C1, 0x00C2, 0x0102, 0x00C4, 0x0139, 0x0106, 0x00C7,
  0x010C, 0x00C9, 0x0118, 0x00CB, 0x011A, 0x00CD, 0x00CE, 0x010E,
  0x0110, 0x0143, 0x0147, 0x00D3, 0x00D4, 0x0150, 0x00D6, 0x00D7,
  0x0158, 0x016E, 0x00DA, 0x0170, 0x00DC, 0x00DD, 0x0162, 0x00DF,
  0x0155, 0x00E1, 0x00E2, 0x0103, 0x00E4, 0x013A, 0x0107, 0x00E7,
  0x010D, 0x00E9, 0x0119, 0x00EB, 0x011B, 0x00ED, 0x00EE, 0x010F,
  0x0111, 0x0144, 0x0148, 0x00F3, 0x00F4, 0x0151, 0x00F6, 0x00F7,
  0x0159, 0x016F, 0x00FA, 0x0171, 0x00FC, 0x00FD, 0x0163, 0x02D9 };

static unsigned short iso8859_3_tophalf[]=
{ 0x00A0, 0x0126, 0x02D8, 0x00A3, 0x00A4, 0x0000, 0x0124, 0x00A7,
  0x00A8, 0x0130, 0x015E, 0x011E, 0x0134, 0x00AD, 0x0000, 0x017B,
  0x00B0, 0x0127, 0x00B2, 0x00B3, 0x00B4, 0x00B5, 0x0125, 0x00B7,
  0x00B8, 0x0131, 0x015F, 0x011F, 0x0135, 0x00BD, 0x0000, 0x017C,
  0x00C0, 0x00C1, 0x00C2, 0x0000, 0x00C4, 0x010A, 0x0108, 0x00C7,
  0x00C8, 0x00C9, 0x00CA, 0x00CB, 0x00CC, 0x00CD, 0x00CE, 0x00CF,
  0x0000, 0x00D1, 0x00D2, 0x00D3, 0x00D4, 0x0120, 0x00D6, 0x00D7,
  0x011C, 0x00D9, 0x00DA, 0x00DB, 0x00DC, 0x016C, 0x015C, 0x00DF,
  0x00E0, 0x00E1, 0x00E2, 0x0000, 0x00E4, 0x010B, 0x0109, 0x00E7,
  0x00E8, 0x00E9, 0x00EA, 0x00EB, 0x00EC, 0x00ED, 0x00EE, 0x00EF,
  0x0000, 0x00F1, 0x00F2, 0x00F3, 0x00F4, 0x0121, 0x00F6, 0x00F7,
  0x011D, 0x00F9, 0x00FA, 0x00FB, 0x00FC, 0x016D, 0x015D, 0x02D9};

static unsigned
iso8859_15_to_unicode(unsigned isocode)
{
    switch(isocode) {
    case 0xA4: return 0x20AC;
    case 0xA6: return 0x0160;
    case 0xA8: return 0x0161;
    case 0xB4: return 0x017D;
    case 0xB8: return 0x017E;
    case 0xBC: return 0x0152;
    case 0xBD: return 0x0153;
    case 0xBE: return 0x0178;
    default: return isocode;
    }
}

static unsigned int
l2c(unsigned int l)
{
    if(l < 0x80) return l;

    switch(localeEncoding) {
    case ISO8859_1: return l;
    case ISO8859_2: 
        if(l < 0xA0) return l; else return iso8859_2_tophalf[l - 0xA0];
    case ISO8859_3:
        if(l < 0xA0) return l; else return iso8859_3_tophalf[l - 0xA0];
    case ISO8859_15:
        return iso8859_15_to_unicode(l);
    default:
        abort();
    }
}

static int 
l2u(char *d, char *s)
{
    unsigned int c;

    c = l2c(*s & 0xFF);
        
    if(c <= 0x7F) {
        *d++ = c;
        return 1;
    } else if(c <= 0x7FF) {
        *d++ = 0xC0 | ((c >> 6) & 0x1F);
        *d++ = 0x80 | (c & 0x3F);
        return 2;
    } else if (c <= 0xFFFF) {
        *d++ = 0xE0 | ((c >> 12) & 0x0F);
        *d++ = 0x80 | ((c >> 6) & 0x3F);
        *d++ = 0x80 | (c & 0x3F);
        return 3;
    } else {
        *d++ = 0xF0 | ((c >> 18) & 0x07);
        *d++ = 0x80 | ((c >> 12) & 0x3F);
        *d++ = 0x80 | ((c >> 6) & 0x3F);
        *d++ = 0x80 | (c & 0x3F);
        return 4;
    }
}

int
parseLocaleEncoding(char *e)
{
    char buf[50];
    int i, j, n;

    n = strlen(e);
    if(n >= 50)
        return ISO8859_1;
    
    i = j = 0;
    while(i < n) {
        if(e[i] != ' ' && e[i] != '-' && e[i] != '_')
            buf[j++] = tolower(e[i]);
        i++;
    }
    buf[j] = '\0';

    i = 0;
    if(memcmp(buf, "iso", 3) == 0)
       i += 3;

    if(strcmp(buf + i, "88591") == 0)
        return ISO8859_1;
    else if(strcmp(buf + i, "88592") == 0)
        return ISO8859_2;
    else if(strcmp(buf + i, "88593") == 0)
        return ISO8859_3;
    else if(strcmp(buf + i, "885915") == 0)
        return ISO8859_15;
    else
        return ISO8859_1;
}
